@@ -12,7 +12,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use std::borrow::Cow;
+use std::cmp::min;
 use std::collections::HashMap;
 use std::pin::Pin;
 use std::str::FromStr;
@@ -22,24 +22,24 @@ use std::task::Poll;
 
 use anyhow::anyhow;
 use async_trait::async_trait;
-use aws_sdk_s3;
-use aws_sdk_s3::Client;
-use aws_smithy_http::body::SdkBody;
-use aws_smithy_http::byte_stream::ByteStream;
+use bytes::BufMut;
 use futures::TryStreamExt;
+use http::header::HeaderName;
 use http::HeaderValue;
+use http::Response;
 use http::StatusCode;
+use hyper::body::HttpBody;
+use hyper::Body;
 use log::debug;
 use log::error;
 use log::info;
 use log::warn;
 use metrics::increment_counter;
 use once_cell::sync::Lazy;
+use reqsign::services::aws::v4::Signer;
+use time::format_description::well_known::Rfc2822;
+use time::OffsetDateTime;
 
-use super::error::parse_get_object_error;
-use super::error::parse_head_object_error;
-use super::error::parse_unexpect_error;
-use super::middleware::DefaultMiddleware;
 use super::object_stream::S3ObjectStream;
 use crate::credential::Credential;
 use crate::error::Error;
@@ -81,13 +81,8 @@ pub struct Builder {
 
     bucket: String,
     credential: Option<Credential>,
-    /// endpoint must be full uri, e.g.
-    /// - <https://s3.amazonaws.com>
-    /// - <http://127.0.0.1:3000>
-    ///
-    /// If user inputs endpoint like "s3.amazonaws.com", we will prepend
-    /// "https://" before it.
     endpoint: Option<String>,
+    region: Option<String>,
 }
 
 impl Builder {
@@ -113,6 +108,12 @@ impl Builder {
         self
     }
 
+    /// endpoint must be full uri, e.g.
+    /// - <https://s3.amazonaws.com>
+    /// - <http://127.0.0.1:3000>
+    ///
+    /// If user inputs endpoint like "s3.amazonaws.com", we will prepend
+    /// "https://" before it.
     pub fn endpoint(&mut self, endpoint: &str) -> &mut Self {
         self.endpoint = if endpoint.is_empty() {
             None
@@ -123,6 +124,110 @@ impl Builder {
         self
     }
 
+    /// region represent the region of this endpoint.
+    ///
+    /// If region is not filled, we will try to detect region via
+    /// RFC-0057: Auto Region
+    pub fn region(&mut self, region: &str) -> &mut Self {
+        self.region = if region.is_empty() {
+            None
+        } else {
+            Some(region.to_string())
+        };
+
+        self
+    }
+
+    // Read RFC-0057: Auto Region for detailed behavior.
+    async fn detect_region(
+        &self,
+        client: &hyper::Client<
+            hyper_tls::HttpsConnector<hyper::client::HttpConnector>,
+            hyper::Body,
+        >,
+        endpoint: &str,
+        bucket: &str,
+        context: &HashMap<String, String>,
+    ) -> Result<(String, String)> {
+        if let Some(region) = &self.region {
+            return Ok((endpoint.to_string(), region.to_string()));
+        }
+
+        let req = hyper::Request::head(format!("{endpoint}/{bucket}"))
+            .body(hyper::Body::empty())
+            .expect("must be valid request");
+        let res = client.request(req).await.map_err(|e| Error::Backend {
+            kind: Kind::BackendConfigurationInvalid,
+            context: context.clone(),
+            source: anyhow::Error::new(e),
+        })?;
+
+        debug!(
+            "auto detect region got response: status {:?}, header: {:?}",
+            res.status(),
+            res.headers()
+        );
+        match res.status() {
+            // The endpoint works, return with not changed endpoint and
+            // default region.
+            StatusCode::OK | StatusCode::FORBIDDEN => {
+                let region = res
+                    .headers()
+                    .get("x-amz-bucket-region")
+                    .unwrap_or(&HeaderValue::from_static("us-east-1"))
+                    .to_str()
+                    .map_err(|e| Error::Backend {
+                        kind: Kind::BackendConfigurationInvalid,
+                        context: context.clone(),
+                        source: anyhow::Error::new(e),
+                    })?
+                    .to_string();
+                Ok((endpoint.to_string(), region))
+            }
+            // The endpoint should move, return with constructed endpoint
+            StatusCode::MOVED_PERMANENTLY => {
+                let region = res
+                    .headers()
+                    .get("x-amz-bucket-region")
+                    .ok_or(Error::Backend {
+                        kind: Kind::BackendConfigurationInvalid,
+                        context: context.clone(),
+                        source: anyhow!("can't detect region automatically, region is empty"),
+                    })?
+                    .to_str()
+                    .map_err(|e| Error::Backend {
+                        kind: Kind::BackendConfigurationInvalid,
+                        context: context.clone(),
+                        source: anyhow::Error::new(e),
+                    })?
+                    .to_string();
+                let template = ENDPOINT_TEMPLATES.get(endpoint).ok_or(Error::Backend {
+                    kind: Kind::BackendConfigurationInvalid,
+                    context: context.clone(),
+                    source: anyhow!(
+                        "can't detect region automatically, no valid endpoint template for {}",
+                        &endpoint
+                    ),
+                })?;
+
+                let endpoint = template.replace("{region}", &region);
+
+                Ok((endpoint, region))
+            }
+            // Unexpected status code
+            code => {
+                return Err(Error::Backend {
+                    kind: Kind::BackendConfigurationInvalid,
+                    context: context.clone(),
+                    source: anyhow!(
+                        "can't detect region automatically, unexpected response: status code {}",
+                        code
+                    ),
+                });
+            }
+        }
+    }
+
     pub async fn finish(&mut self) -> Result<Arc<dyn Accessor>> {
         info!("backend build started: {:?}", &self);
 
@@ -153,120 +258,28 @@ impl Builder {
         }?;
         debug!("backend use bucket {}", &bucket);
 
+        // Setup error context so that we don't need to construct many times.
+        let mut context: HashMap<String, String> =
+            HashMap::from([("bucket".to_string(), bucket.to_string())]);
+
+        let client = hyper::Client::builder().build(hyper_tls::HttpsConnector::new());
+
         let endpoint = match &self.endpoint {
             Some(endpoint) => endpoint,
             None => "https://s3.amazonaws.com",
         };
-        debug!("backend use endpoint {} to detect region", &endpoint);
 
-        // Setup error context so that we don't need to construct many times.
-        let mut context: HashMap<String, String> = HashMap::from([
-            ("endpoint".to_string(), endpoint.to_string()),
-            ("bucket".to_string(), bucket.to_string()),
-        ]);
-
-        let hc = reqwest::Client::new();
-        let res = hc
-            .head(format!("{endpoint}/{bucket}"))
-            .send()
-            .await
-            .map_err(|e| Error::Backend {
-                kind: Kind::BackendConfigurationInvalid,
-                context: context.clone(),
-                source: anyhow::Error::new(e),
-            })?;
-        // Read RFC-0057: Auto Region for detailed behavior.
-        let (endpoint, region) = match res.status() {
-            // The endpoint works, return with not changed endpoint and
-            // default region.
-            StatusCode::OK | StatusCode::FORBIDDEN => {
-                let region = res
-                    .headers()
-                    .get("x-amz-bucket-region")
-                    .unwrap_or(&HeaderValue::from_static("us-east-1"))
-                    .to_str()
-                    .map_err(|e| Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow::Error::new(e),
-                    })?
-                    .to_string();
-                (endpoint.to_string(), region)
-            }
-            // The endpoint should move, return with constructed endpoint
-            StatusCode::MOVED_PERMANENTLY => {
-                let region = res
-                    .headers()
-                    .get("x-amz-bucket-region")
-                    .ok_or(Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow!("can't detect region automatically, region is empty"),
-                    })?
-                    .to_str()
-                    .map_err(|e| Error::Backend {
-                        kind: Kind::BackendConfigurationInvalid,
-                        context: context.clone(),
-                        source: anyhow::Error::new(e),
-                    })?
-                    .to_string();
-                let template = ENDPOINT_TEMPLATES.get(endpoint).ok_or(Error::Backend {
-                    kind: Kind::BackendConfigurationInvalid,
-                    context: context.clone(),
-                    source: anyhow!(
-                        "can't detect region automatically, no valid endpoint template for {}",
-                        &endpoint
-                    ),
-                })?;
-
-                let endpoint = template.replace("{region}", &region);
-
-                (endpoint, region)
-            }
-            // Unexpected status code
-            code => {
-                return Err(Error::Backend {
-                    kind: Kind::BackendConfigurationInvalid,
-                    context: context.clone(),
-                    source: anyhow!(
-                        "can't detect region automatically, unexpected response: status code {}",
-                        code
-                    ),
-                });
-            }
-        };
+        let (endpoint, region) = self
+            .detect_region(&client, endpoint, bucket, &context)
+            .await?;
+        context.insert("endpoint".to_string(), endpoint.clone());
+        context.insert("region".to_string(), region.clone());
         debug!("backend use endpoint: {}, region: {}", &endpoint, &region);
 
-        // Config Loader will load config from environment.
-        //
-        // We will take user's input first if any. If there is no user input, we
-        // will fallback to the aws default load chain like the following:
-        //
-        // - Environment variables: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, and AWS_REGION
-        // - The default credentials files located in ~/.aws/config and ~/.aws/credentials (location can vary per platform)
-        // - Web Identity Token credentials from the environment or container (including EKS)
-        // - ECS Container Credentials (IAM roles for tasks)
-        // - EC2 Instance Metadata Service (IAM Roles attached to instance)
-        //
-        // Please keep in mind that the config loader only detect region and credentials.
-        let cfg_loader = aws_config::ConfigLoader::default();
-        let mut cfg = aws_sdk_s3::config::Builder::from(&cfg_loader.load().await);
-
-        {
-            // Set region.
-            cfg = cfg.region(aws_sdk_s3::Region::new(Cow::from(region.clone())));
-        }
-
-        {
-            // Set endpoint
-            let uri = http::Uri::from_str(&endpoint).map_err(|e| Error::Backend {
-                kind: Kind::BackendConfigurationInvalid,
-                context: context.clone(),
-                source: anyhow::Error::from(e),
-            })?;
-
-            cfg = cfg.endpoint_resolver(aws_sdk_s3::Endpoint::immutable(uri));
-        }
+        let mut signer_builder = reqsign::services::aws::v4::Signer::builder();
+        signer_builder.service("s3");
+        signer_builder.region(&region);
+        signer_builder.allow_anonymous();
 
         if let Some(cred) = &self.credential {
             context.insert("credential".to_string(), "*".to_string());
@@ -275,11 +288,8 @@ impl Builder {
                     access_key_id,
                     secret_access_key,
                 } => {
-                    cfg = cfg.credentials_provider(aws_sdk_s3::Credentials::from_keys(
-                        access_key_id,
-                        secret_access_key,
-                        None,
-                    ));
+                    signer_builder.access_key(access_key_id);
+                    signer_builder.secret_key(secret_access_key);
                 }
                 // We don't need to do anything if user tries to read credential from env.
                 Credential::Plain => {
@@ -295,22 +305,15 @@ impl Builder {
             }
         }
 
-        let hyper_connector = aws_smithy_client::hyper_ext::Adapter::builder()
-            .build(aws_smithy_client::conns::https());
-
-        let aws_client = aws_smithy_client::Builder::new()
-            .connector(hyper_connector)
-            .middleware(aws_smithy_client::erase::DynMiddleware::new(
-                DefaultMiddleware::new(),
-            ))
-            .default_async_sleep()
-            .build();
+        let signer = signer_builder.build().await?;
 
         info!("backend build finished: {:?}", &self);
         Ok(Arc::new(Backend {
             root,
+            endpoint,
+            signer: Arc::new(signer),
             bucket: self.bucket.clone(),
-            client: aws_sdk_s3::Client::with_config(aws_client.into_dyn(), cfg.build()),
+            client,
         }))
     }
 }
@@ -319,8 +322,9 @@ impl Builder {
 #[derive(Debug, Clone)]
 pub struct Backend {
     bucket: String,
-
-    client: aws_sdk_s3::Client,
+    endpoint: String,
+    signer: Arc<Signer>,
+    client: hyper::Client<hyper_tls::HttpsConnector<hyper::client::HttpConnector>, hyper::Body>,
     // root will be "/" or "/abc/"
     root: String,
 }
@@ -330,10 +334,6 @@ impl Backend {
         Builder::default()
     }
 
-    pub(crate) fn inner(&self) -> Client {
-        self.client.clone()
-    }
-
     // normalize_path removes all internal `//` inside path.
     pub(crate) fn normalize_path(path: &str) -> String {
         let has_trailing = path.ends_with('/');
@@ -388,52 +388,33 @@ impl Accessor for Backend {
             &p, args.offset, args.size
         );
 
-        let mut req = self
-            .client
-            .get_object()
-            .bucket(&self.bucket.clone())
-            .key(&p);
+        let resp = self.get_object(&p, args.offset, args.size).await?;
 
-        if args.offset.is_some() || args.size.is_some() {
-            req = req.range(HeaderRange::new(args.offset, args.size).to_string());
+        match resp.status() {
+            StatusCode::OK | StatusCode::PARTIAL_CONTENT => {
+                info!(
+                    "object {} reader created: offset {:?}, size {:?}",
+                    &p, args.offset, args.size
+                );
+
+                Ok(Box::new(ByteStream(resp).into_async_read()))
+            }
+            _ => Err(parse_error_response(resp, "read", &p).await),
         }
-
-        let resp = req.send().await.map_err(|e| {
-            let e = parse_get_object_error(e, "read", &p);
-            error!("object {} get_object: {:?}", &p, e);
-            e
-        })?;
-
-        info!(
-            "object {} reader created: offset {:?}, size {:?}",
-            &p, args.offset, args.size
-        );
-        Ok(Box::new(S3ByteStream(resp.body).into_async_read()))
     }
 
     async fn write(&self, r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
         let p = self.get_abs_path(&args.path);
         info!("object {} write start: size {}", &p, args.size);
 
-        let _ = self
-            .client
-            .put_object()
-            .bucket(&self.bucket.clone())
-            .key(&p)
-            .content_length(args.size as i64)
-            .body(ByteStream::from(SdkBody::from(
-                hyper::body::Body::wrap_stream(ReaderStream::new(r)),
-            )))
-            .send()
-            .await
-            .map_err(|e| {
-                let e = parse_unexpect_error(e, "write", &p);
-                error!("object {} put_object: {:?}", &p, e);
-                e
-            })?;
-
-        info!("object {} write finished: size {:?}", &p, args.size);
-        Ok(args.size as usize)
+        let resp = self.put_object(&p, r, args.size).await?;
+        match resp.status() {
+            StatusCode::CREATED | StatusCode::OK => {
+                info!("object {} write finished: size {:?}", &p, args.size);
+                Ok(args.size as usize)
+            }
+            _ => Err(parse_error_response(resp, "write", &p).await),
+        }
     }
 
     async fn stat(&self, args: &OpStat) -> Result<Metadata> {
@@ -454,20 +435,35 @@ impl Accessor for Backend {
             return Ok(m);
         }
 
-        let meta = self
-            .client
-            .head_object()
-            .bucket(&self.bucket.clone())
-            .key(&p)
-            .send()
-            .await
-            .map_err(|e| parse_head_object_error(e, "stat", &p));
+        let resp = self.head_object(&p).await?;
 
-        match meta {
-            Ok(meta) => {
+        match resp.status() {
+            StatusCode::OK => {
                 let mut m = Metadata::default();
                 m.set_path(&args.path);
-                m.set_content_length(meta.content_length as u64);
+
+                // Parse content_length
+                if let Some(v) = resp.headers().get(http::header::CONTENT_LENGTH) {
+                    let v =
+                        u64::from_str(v.to_str().expect("header must not contain non-ascii value"))
+                            .expect("content length header must contain valid length");
+
+                    m.set_content_length(v);
+                }
+
+                // Parse content_md5
+                if let Some(v) = resp.headers().get(HeaderName::from_static("content-md5")) {
+                    let v = v.to_str().expect("header must not contain non-ascii value");
+                    m.set_content_md5(v);
+                }
+
+                // Parse last_modified
+                if let Some(v) = resp.headers().get(http::header::LAST_MODIFIED) {
+                    let v = v.to_str().expect("header must not contain non-ascii value");
+                    let t =
+                        OffsetDateTime::parse(v, &Rfc2822).expect("must contain valid time format");
+                    m.set_last_modified(t.into());
+                }
 
                 if p.ends_with('/') {
                     m.set_mode(ObjectMode::DIR);
@@ -477,12 +473,10 @@ impl Accessor for Backend {
 
                 m.set_complete();
 
-                info!("object {} stat finished", &p);
+                info!("object {} stat finished: {:?}", &p, m);
                 Ok(m)
             }
-            // Always returns empty dir object if path is endswith "/" and we got an
-            // ObjectNotExist error.
-            Err(e) if (e.kind() == Kind::ObjectNotExist && p.ends_with('/')) => {
+            StatusCode::NOT_FOUND if p.ends_with('/') => {
                 let mut m = Metadata::default();
                 m.set_path(&args.path);
                 m.set_content_length(0);
@@ -492,10 +486,7 @@ impl Accessor for Backend {
                 info!("object {} stat finished", &p);
                 Ok(m)
             }
-            Err(e) => {
-                error!("object {} head_object: {:?}", &p, e);
-                Err(e)
-            }
+            _ => Err(parse_error_response(resp, "stat", &p).await),
         }
     }
 
@@ -505,17 +496,15 @@ impl Accessor for Backend {
         let p = self.get_abs_path(&args.path);
         info!("object {} delete start", &p);
 
-        let _ = self
-            .client
-            .delete_object()
-            .bucket(&self.bucket.clone())
-            .key(&p)
-            .send()
-            .await
-            .map_err(|e| parse_unexpect_error(e, "delete", &p))?;
+        let resp = self.delete_object(&p).await?;
 
-        info!("object {} delete finished", &p);
-        Ok(())
+        match resp.status() {
+            StatusCode::NO_CONTENT => {
+                info!("object {} delete finished", &p);
+                Ok(())
+            }
+            _ => Err(parse_error_response(resp, "delete", &p).await),
+        }
     }
 
     async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
@@ -528,35 +517,185 @@ impl Accessor for Backend {
         }
         info!("object {} list start", &path);
 
-        Ok(Box::new(S3ObjectStream::new(
-            self.clone(),
-            self.bucket.clone(),
-            path,
-        )))
+        Ok(Box::new(S3ObjectStream::new(self.clone(), path)))
     }
 }
 
-struct S3ByteStream(aws_smithy_http::byte_stream::ByteStream);
+impl Backend {
+    pub(crate) async fn get_object(
+        &self,
+        path: &str,
+        offset: Option<u64>,
+        size: Option<u64>,
+    ) -> Result<hyper::Response<hyper::Body>> {
+        let mut req = hyper::Request::get(&format!("{}/{}/{}", self.endpoint, self.bucket, path));
 
-impl futures::Stream for S3ByteStream {
-    type Item = std::result::Result<bytes::Bytes, std::io::Error>;
+        if offset.is_some() || size.is_some() {
+            req = req.header(
+                http::header::RANGE,
+                HeaderRange::new(offset, size).to_string(),
+            );
+        }
+        let mut req = req
+            .body(hyper::Body::empty())
+            .expect("must be valid request");
+
+        self.signer.sign(&mut req).await.expect("sign must success");
+
+        self.client.request(req).await.map_err(|e| {
+            error!("object {} get_object: {:?}", path, e);
+            Error::Object {
+                kind: Kind::Unexpected,
+                op: "read",
+                path: path.to_string(),
+                source: anyhow::Error::from(e),
+            }
+        })
+    }
+
+    pub(crate) async fn put_object(
+        &self,
+        path: &str,
+        r: BoxedAsyncReader,
+        size: u64,
+    ) -> Result<hyper::Response<hyper::Body>> {
+        let mut req = hyper::Request::put(&format!("{}/{}/{}", self.endpoint, self.bucket, path));
+
+        // Set content length.
+        req = req.header(http::header::CONTENT_LENGTH, size.to_string());
+
+        // Set body
+        let mut req = req
+            .body(hyper::body::Body::wrap_stream(ReaderStream::new(r)))
+            .expect("must be valid request");
+
+        self.signer.sign(&mut req).await.expect("sign must success");
+
+        self.client.request(req).await.map_err(|e| {
+            error!("object {} put_object: {:?}", path, e);
+            Error::Object {
+                kind: Kind::Unexpected,
+                op: "write",
+                path: path.to_string(),
+                source: anyhow::Error::from(e),
+            }
+        })
+    }
+
+    pub(crate) async fn head_object(&self, path: &str) -> Result<hyper::Response<hyper::Body>> {
+        let mut req = hyper::Request::head(&format!("{}/{}/{}", self.endpoint, self.bucket, path))
+            .body(hyper::Body::empty())
+            .expect("must be valid request");
+
+        self.signer.sign(&mut req).await.expect("sign must success");
+
+        self.client.request(req).await.map_err(|e| {
+            error!("object {} head_object: {:?}", path, e);
+            Error::Object {
+                kind: Kind::Unexpected,
+                op: "stat",
+                path: path.to_string(),
+                source: anyhow::Error::from(e),
+            }
+        })
+    }
+
+    pub(crate) async fn delete_object(&self, path: &str) -> Result<hyper::Response<hyper::Body>> {
+        let mut req =
+            hyper::Request::delete(&format!("{}/{}/{}", self.endpoint, self.bucket, path))
+                .body(hyper::Body::empty())
+                .expect("must be valid request");
+
+        self.signer.sign(&mut req).await.expect("sign must success");
+
+        self.client.request(req).await.map_err(|e| {
+            error!("object {} delete_object: {:?}", path, e);
+            Error::Object {
+                kind: Kind::Unexpected,
+                op: "delete",
+                path: path.to_string(),
+                source: anyhow::Error::from(e),
+            }
+        })
+    }
+
+    pub(crate) async fn list_object(
+        &self,
+        path: &str,
+        continuation_token: &str,
+    ) -> Result<hyper::Response<hyper::Body>> {
+        let mut uri = format!(
+            "{}/{}?list-type=2&delimiter=/&prefix={}",
+            self.endpoint, self.bucket, path
+        );
+        if !continuation_token.is_empty() {
+            uri.push_str(&format!("&continuation-token={}", continuation_token))
+        }
+
+        let mut req = hyper::Request::get(uri)
+            .body(hyper::Body::empty())
+            .expect("must be valid request");
+
+        self.signer.sign(&mut req).await.expect("sign must success");
+
+        self.client.request(req).await.map_err(|e| {
+            error!("object {} list_object: {:?}", path, e);
+            Error::Object {
+                kind: Kind::Unexpected,
+                op: "list",
+                path: path.to_string(),
+                source: anyhow::Error::from(e),
+            }
+        })
+    }
+}
+
+struct ByteStream(hyper::Response<hyper::Body>);
+
+impl futures::Stream for ByteStream {
+    type Item = std::io::Result<bytes::Bytes>;
 
-    /// ## TODO
-    ///
-    /// This hack is ugly, we should find a better way to do this.
-    ///
-    /// The problem is `into_async_read` requires the stream returning
-    /// `std::io::Error`, the the `ByteStream` returns
-    /// `aws_smithy_http::byte_stream::Error` instead.
-    ///
-    /// I don't know why aws sdk should wrap the error into their own type...
     fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
-        Pin::new(&mut self.0)
+        Pin::new(self.0.body_mut())
             .poll_next(cx)
-            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))
+            .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e.to_string()))
     }
+}
 
-    fn size_hint(&self) -> (usize, Option<usize>) {
-        self.0.size_hint()
+// Read and decode whole error response.
+async fn parse_error_response(resp: Response<Body>, op: &'static str, path: &str) -> Error {
+    let (part, mut body) = resp.into_parts();
+    let kind = match part.status {
+        StatusCode::NOT_FOUND => Kind::ObjectNotExist,
+        StatusCode::FORBIDDEN => Kind::ObjectPermissionDenied,
+        _ => Kind::Unexpected,
+    };
+
+    // Only read 4KiB from the response to avoid broken services.
+    let mut bs = Vec::new();
+    let mut limit = 4 * 1024;
+
+    while let Some(b) = body.data().await {
+        match b {
+            Ok(b) => {
+                bs.put_slice(&b[..min(b.len(), limit)]);
+                limit -= b.len();
+                if limit == 0 {
+                    break;
+                }
+            }
+            Err(e) => return Error::Unexpected(anyhow!("parse error response parse: {:?}", e)),
+        }
+    }
+
+    Error::Object {
+        kind,
+        op,
+        path: path.to_string(),
+        source: anyhow!(
+            "response part: {:?}, body: {:?}",
+            part,
+            String::from_utf8_lossy(&bs)
+        ),
     }
 }
